'''
Tested
'''
from .utils import *
import os

class FASTQC():


    def check(self,nockeck=False):
        #return True,''
        if not toolcheck('fastqc --help'):
            return False,'Fastqc command not found'
        if os.path.exists('Fastqc'):
            if nockeck:
                print('Fastqc file or dir exists! But --nockeck enabled, so continue running')
            else:
                return False,'Fastqc file or dir exists'
        else:
            os.mkdir('Fastqc')
        return True,''

    def setpath(self,path):
        self.path = path+'Fastqc'

    def run(self,filename):
        pshell=Pshell('fastqc -o '+self.path+' '+filename)
        pshell.process()

    
Fastqc = FASTQC()

if __name__=="__main__":
    a = Fastqc()
    #print(a.check())
    a.setpath('./')
    a.run('../trimtest/SRR1248444_1.1.1.1.fastq')



import gzip
import pysam
from .utils import Pshell
import os

def pair_valid(chr1,chr2,strand1,strand2,pos1,pos2):
    if chr1!=chr2:
        return False
    if strand1=='++' and strand2!='+-':
        return False
    if strand1=='-+' and strand2!='--':
        return False
    if strand1=='+-' and strand2!='++':
        return False
    if strand1=='--' and strand2!='-+':
        return False
    if abs(int(pos1)-int(pos2))<=1000:
        return True
    return False
 

def writefiles(UnmappedReads, outputname):

    filecontent = []
    for readsname in UnmappedReads:
        link = UnmappedReads[readsname]
        filecontent.append('@'+readsname+'\n'+link[0]+'\n+\n'+link[1]+'\n')
    
    with open(outputname,'a') as f:
        f.writelines(filecontent)
    


def pairend_remap(bam_filename,fastq_filename,output_name,refpath,output_prefix, multihits_num=10):
    '''
        Args:
            bam_filename: string, name of combined bam file
            fastq_filename: string[], list of names of fastq files
        Output:
            1. 
    '''

    #with open(outputname+".sam") as sam:
    #second column in sam file: 64, mate 1; 128, mate 2;
    #    samlines = sam.readlines()
    output_name_fq = output_name + '_multi.fq'
    output_name_bam = output_name + '_multi.bam'
    bamfile = pysam.AlignmentFile(bam_filename, "rb")
    set_sam = {}
    fq_file_num = len(fastq_filename)
    for line in bamfile:
        temp = line.to_dict()
        name = temp['name']
        if temp['next_ref_name']!='*':
            continue
        clip_from = -1
        if '_' in name:
            pos = name.find('_')
            if name[pos+3]=='0':
                clip_from = 1
            else:
                clip_from = 2
            name = name[:pos]
        flag = int(temp['flag'])
        m1 = (flag & 64)
        m2 = (flag & 128)
        if m1>0: mate = 1
        elif m2>0: mate = 2
        else: mate = 0
        if clip_from>0 and fq_file_num>1:
            mate = clip_from
        if name in set_sam: set_sam[name]=3
        else: set_sam[name]=mate 
    bamfile.close()
    
    UnmappedReads = {}
    o=0

    for filename in fastq_filename:
        o+=1
        gzmark=False
        if filename.endswith('.gz'):
            f = gzip.open(filename)
            gzmark=True
        else:
            f = open(filename)


        if f:
            while 1:
                if gzmark:
                    line1 = f.readline().decode()
                else:
                    line1 = f.readline()
                if not line1:
                    break
                if gzmark:
                    line2 = f.readline().decode().strip()
                    line3 = f.readline().decode()
                    line4 = f.readline().decode().strip()
                else:
                    line2 = f.readline().strip()
                    line3 = f.readline()
                    line4 = f.readline().strip()
                line1 = line1.strip().split()
                read_name = line1[0][1:]
                pair_file_rank = o
                
                if '/' in read_name: 
                    split_pos=0
                    split_pos = read_name.rfind('/')
                    read_name = read_name[:split_pos]

                if (read_name in set_sam):
                    if set_sam[read_name]==0 or set_sam[read_name]==3: continue
                    if set_sam[read_name]==pair_file_rank: continue
                    if pair_file_rank<0: continue
                else:
                    continue

                #if fq_file_num>1: read_name+='_'+str(pair_file_rank)
                UnmappedReads[read_name]=[line2,line4]
                if len(UnmappedReads)>1000000:
                    writefiles(UnmappedReads,output_name_fq)
                    UnmappedReads={}
        f.close()            

    if len(UnmappedReads)>0:
        writefiles(UnmappedReads,output_name_fq)
    #print('finish')
    del UnmappedReads

    command='bsmap -a '+output_name_fq+' -d '+refpath+'  -o '+output_name_bam+' -w '+str(multihits_num)+' -S 123 -n 1 -r 2 1>>temp_log 2>>'+output_prefix+'_multi_log.txt'
    multiple_mapping = Pshell(command)
    multiple_mapping.process()
    if os.path.exists(output_name_fq):
        os.remove(output_name_fq)
    return output_name_bam

def bam_merge(ori_bam, multi_bam):
    '''
        Args:
            ori_bam: string, name of the original *_combined.bam
            multi_bam: string, name of the new multiple mapping bam
    '''
    
    samfile = pysam.AlignmentFile(multi_bam, "rb") 
    multi_dict = {}
    for line in samfile:
        temp = line.to_dict()
        chr = temp['ref_name']
        pos = temp['ref_pos']
        strand=''
        for tag in temp['tags']:
            if 'ZS' in tag:
                strand = tag[-2:]
        name = temp['name'][:-2]
        if name in multi_dict:
            multi_dict[name].append([chr,pos,strand])
        else:
            multi_dict[name] = [[chr,pos,strand]]

    samfile.close()
    potential_marker = {}
    samfile = pysam.AlignmentFile(ori_bam, "rb")
    for line in samfile:
        temp = line.to_dict()
        if temp['next_ref_name']!='*':
            continue
        name = temp['name']
        if name not in multi_dict:
            continue
        chr = temp['ref_name']
        pos = int(temp['ref_pos'])
        strand=''
        for tag in temp['tags']:
            if 'ZS' in tag:
                strand = tag[-2:]
        
        candidates = multi_dict[name]
        for c in candidates:
            if chr==c[0]:
                if strand=='++' and c[2]!='+-':
                    continue
                if strand=='-+' and c[2]!='--':
                    continue
                if strand=='+-' and c[2]!='++':
                    continue
                if strand=='--' and c[2]!='-+':
                    continue
                if abs(pos-int(c[1]))<=1000:
                    if name in potential_marker:
                        potential_marker[name].append(c)
                    else:
                        potential_marker[name] = [c]
                    
    samfile.close()
        
    multifile = pysam.AlignmentFile(multi_bam, "rb") 
    output_file = pysam.AlignmentFile(multi_bam+'.select.bam','wb',template=multifile)
    for line in multifile:
        temp = line.to_dict()
        name = temp['name']
        #mate = int(name[-1])
        #name = name[:-2]
        if name not in potential_marker:
            continue
        mate = int(name[-1])
        if mate==1:
            flag = 64
        else:
            flag = 128
        #line.flag |= 
        chr = temp['ref_name']
        pos = temp['ref_pos']
        for c in potential_marker:
            if chr==c[0] and pos==c[1]:
                output_file.write(line)
    multifile.close()
    output_file.close()
    return multi_bam+'.select.bam'

def libis_filter(ori_bam, split_bam):
    '''
        Args:
            ori_bam: string, name of the original *_combined.bam
            multi_bam: string, name of the new multiple mapping bam
    '''
    
    samfile = pysam.AlignmentFile(ori_bam, "rb") 
    ori_dict = {}
    for line in samfile:
        temp = line.to_dict()
        if temp['next_ref_name']!='*':
            continue
        chr = temp['ref_name']
        pos = temp['ref_pos']
        strand=''
        for tag in temp['tags']:
            if 'ZS' in tag:
                strand = tag[-2:]
        name = temp['name'][:-2]
        ori_dict[name] = [chr,pos,strand]

    samfile.close()
    split_dict = {}
    candidates = {}
    samfile = pysam.AlignmentFile(split_bam, "rb")
    output_file = pysam.AlignmentFile(split_bam+'.filter.bam','wb',template=samfile)
    number = 0
    for line in samfile:
        temp = line.to_dict()
        name = temp['name']
        #print(name)
        _pos = name.find('_')
        #print(pos)
        #print(name[:pos])
        clip_from = -1
        if name[_pos+3]=='1':
            clip_from = 1
        else:
            clip_from = 2
        chr = temp['ref_name']
        pos = temp['ref_pos']
        strand = ''
        for tag in temp['tags']:
            if 'ZS' in tag:
                strand = tag[-2:]
        name = name[:_pos]
        if name in ori_dict:
            chr1, pos1, strand1 = ori_dict[name]
            if pair_valid(chr,chr1,strand,strand1,pos,pos1):
                output_file.write(line)
                #number += 1
        else:
            if name not in split_dict:
                split_dict[name] = [[clip_from, chr, pos, strand]]
            else:
                split_dict[name].append([clip_from, chr, pos, strand])
    samfile.close()
    for name, c in split_dict.items():
        if len(c)<2: continue
        clip_from = 0
        for cc in c:
            clip_from = clip_from | cc[0]
        if clip_from!=3: continue
        #candidate.append(name)
        mark = False
        cand = set()
        for i in range(len(c)):
            for j in range(i+1,len(c)):
                if c[i][0] | c[j][0]==3 and pair_valid(c[i][1],c[j][1],c[i][3],c[j][3],c[i][2],c[j][2]):
                    cand.add((c[j][1],c[j][2],c[j][3]))
                    cand.add((c[i][1],c[i][2],c[i][3]))
                    mark = True
        if mark:
            candidates[name] = list(cand)
                    #number += 1
    #print(number)
        
    splitfile = pysam.AlignmentFile(split_bam, "rb") 
    for line in splitfile:
        temp = line.to_dict()
        name = temp['name']
        #mate = int(name[-1])
        #name = name[:-2]
        _pos = name.find('_')
        name = name[:_pos]
        if name not in candidates:
            continue
        chr = temp['ref_name']
        pos = temp['ref_pos']
        strand=''
        for tag in temp['tags']:
            if 'ZS' in tag:
                strand = tag[-2:]
        for c in candidates[name]:
            if chr==c[0] and pos==c[1] and strand==c[2]:
                output_file.write(line)
    splitfile.close()
    output_file.close()
    return split_bam+'.filter.bam'
    

if __name__=="__main__":
    import sys
    refpath = '/fdata/scratch/lijin.bio/resource/genome/hg38/hg38.fa'
    #bam = sys.argv[1]#'H1_1A.bam'
    #fq = sys.argv[2:4]#['H1_1A_CKDL190142710-1a_HMFVMDSXX_L2_1_val_1.fq.gz','H1_1A_CKDL190142710-1a_HMFVMDSXX_L2_2_val_2.fq.gz']
    #pairend_remap(bam,fq,bam[:bam.find('.')]+'_multi',refpath)   
    #bam_merge('H1_1D.bam','H1_1D_multi.bam')
    ori_bam, split_bam = sys.argv[1], sys.argv[2]
    libis_filter(ori_bam, split_bam)



'''
Tested
This is a wapper for MOABS in this pipeline. It contains commands for BSMAP, bam sort, bam deduplicate, methylation calling.
'''

from .utils import *
from .clipmode import clipmode
import os

class MCALL():

    def check(self,nocheck=False):
        #return True,''
        if not toolcheck('mcall'):
            return False,'Mcall not found!'
        if os.path.exists('BED_FILE'):
            if nocheck:
                print('"BED_FILE" exists! But --nocheck enabled, so continue running.')
            else:
                return False,'"BED_FILE" exists! Please delete "BED_FILE"'
        else:
            os.mkdir("BED_FILE")
        return True,''

    def setpath(self,path):
        self.path = path+'BED_FILE/'

    def setparam(self,param):
        mcallparam=None
        if 'mcall' in param:
            mcallparam=param['mcall']
        self.refpath = param['ref']
        self.extraparam=''

    def run(self,name):
        '''
        4 muted 37 41 46 49
        '''
        cmd='mcall -m '+name+' -r '+self.refpath+' -p 8 1>>'+self.path+'log 2>>'+self.path+'err'
        p = Pshell(cmd)
        p.process()
        generatedfile=['.G.bed','.HG.bed','_stat.txt']
        newname=[]
        for g in generatedfile:
            os.rename(name+g,self.path+name[name.rfind('/')+1:]+g)
            newname.append(self.path+name[name.rfind('/')+1:]+g)
       
        cmd="awk -v OFS='\t' '{if (NR!=1) print $1,$2,$3,$4}' "+newname[0]+'> '+newname[0]+'.short.bed'
        p.change(cmd)
        p.process()
        cmd="rm mSuite.G.bed"
        p.change(cmd)
        p.process()

        return newname[0],newname[2]

        



class BSMAP():

    def __init__(self):
        self.samtools_version = 1.1#samtoolsversion()
        self.has_unsorted = False
    
    def samtools_sort(self,p, inputfile, outputfile):
        if self.samtools_version<=1.3:
            p.change('samtools sort -f -@ 4 '+inputfile+' '+outputfile)
            p.process()
        else:
            p.change('')
            p.process()
    
    def check(self,nocheck=False):
        #return True,''
        if not toolcheck('bsmap -h'):
            return False,'BSMAP not found!'
        if os.path.exists('BAM_FILE'):
            if nocheck:
                print('"BAM_FILE" exists! But --nocheck enabled, so continue running.')
            else:
                return False,'"BAM_FILE" exists! Please delete "BAM_FILE"'
        else:
            os.mkdir("BAM_FILE")
        return True,''
   
    def setpath(self,path):
        self.path = path+'BAM_FILE/'

    def setparam(self,param):
        self.extraparam=''
        bsmapparam=None
        if 'bsmap' in param:
            bsmapparam = param['bsmap']
        '''
        Once I'm ready to add all parameters for bsmap and mcall (After finish the dictionary 
        to parameter function, I will delete variable named ***param )
        '''
        self.refpath = param['ref']

    def normalmode(self,file,given_label,threads,param={}):
        '''
        3 muted 96 98 100
        '''
        #f = file.strip().split()
        f = file
        purename = given_label#RemoveFastqExtension(f[0][f[0].rfind('/')+1:])
        name = self.path+purename+'.bam'
        logname = self.path+purename+'.record'
        if len(f)==1:
            cmd = 'bsmap -a '+f[0]+' -d '+self.refpath+' -o '+name+' -S 123 -n 1 -p ' + threads + ' 1>>BAM_FILE/bsmap_log 2>'+logname
        else:
            cmd = 'bsmap -a '+f[0]+' -b '+f[1]+' -d '+self.refpath+' -o '+name+' -S 123 -n 1 -p ' + threads + ' 1>>BAM_FILE/bsmap_log 2>'+logname
        p = Pshell(cmd)
        p.process()
        p.change('samtools sort -f -@ 4 '+name+' '+name+'.sorted.bam')
        p.process()
        p.change('mv '+name+'.sorted.bam '+name)
        p.process()
        return name,logname

    def hasUnsortMode(self):
        p = Pshell('bsmap')
        help_mes = p.process_with_err()
        if '-U' in help_mes:
            self.has_unsorted = True

    def clipping(self,filenames, param, given_bam_file,given_label):
        '''
        I should return a bam file name and a log file name here
        '''
        self.hasUnsortMode()
        newname,log = clipmode(filenames,param, given_bam_file,given_label)
        return newname,log


Mcall, Bsmap = MCALL(), BSMAP()


if __name__=="__main__":
    #mcall = Mcall()
    #print(mcall.check())
    #bsmap = Bsmap()
    #bsmap.setparam({'ref':'/data/dsun/ref/humanigenome/hg19.fa'})
    #print(bsmap.check())
    #bsmap.setpath('./')
    #bsmap.normalmode(['Trim/head.fq'])
    mcall = Mcall()
    mcall.setpath('./')
    mcall.setparam({'ref':'/data/dsun/ref/humanigenome/hg19.fa'})
    mcall.run('BAM_FILE/head_combine.bam')



import sys
import multiprocessing
import subprocess
import os
import time
from datetime import datetime

timer = str(int(time.time()))


class Pshell():

    def __init__(self,command):
        self.command = command
        self.out = ""

    def change(self,command):
        self.command = command

    def getcommand(self,command):
        return command

    def process(self):
        t = subprocess.Popen(self.command,shell=True,stdout = subprocess.PIPE,stderr = subprocess.PIPE)
        self.out = ''#t.stdout.read().decode()
        self.err = t.stderr.read().decode()
        with open('LiBis_command_log_'+timer+'.txt','a') as f: 
            f.write(str(datetime.now())+'\tcommand: '+self.command+'\n')
            f.write('Output:\n')
            f.write(self.err+'\n\n')

    def process_with_log(self):
        t = subprocess.Popen(self.command,shell=True,stdout = subprocess.PIPE,stderr = subprocess.PIPE)
        out = t.stdout.read().decode()
        return out

    def process_with_err(self):
        t = subprocess.Popen(self.command,shell=True,stdout = subprocess.PIPE,stderr = subprocess.PIPE)
        err = t.stderr.read().decode()
        return err

    def get_result(self):
        return self.out

class reads():

    def __init__(self,chromosome,startpos,order,mismatch):
       # self.fqname = fqname
        self.chromosome = chromosome
        self.startpos = startpos
        self.sum = 0
        self.order = order
        self.chain = 0
        self.mismatch = mismatch

    def __str__(self):
        print(self.__dict__)

    def equal(self,rreads):
        if (rreads.getChrom()==self.chromosome)and(rreads.getStartPos()==self.startpos): return True
        return False

    def getChrom(self):
        return self.chromosome

    def getOrder(self):
        return self.order

    def getSum(self):
        return self.sum

    def getStartPos(self):
        return self.startpos

    def join(self,tail,mis):
        self.mismatch = self.mismatch + mis
        self.sum = tail.getOrder()-self.getOrder()

    def getChain(self):
        return self.chain

    def getMismatch(self):
        return self.mismatch

    def canjoin(self,tail,step,xx,binsize):
        if (tail.getChrom()==self.chromosome):
            if (tail.getStartPos() == self.startpos + step*(self.order-tail.getOrder()) or 
            tail.getStartPos() == self.startpos - step*(self.order-tail.getOrder())):
            #if tail.getStartPos()>self.startpos: self.chain = 0
            #else: self.chain = 1
                return True
            if xx!=binsize:
                if xx % step == abs(self.startpos-tail.getStartPos())%step:
                    return True
        return False


def readsjoin(head,tail,step,xx,binsize):
    
    if (head[0]==tail[0]):
        if not overlap(head,tail,step,binsize): return False
        if (tail[1]==head[1]+step*(head[2]-tail[2]) or  tail[1]==head[1]-step*(head[2]-tail[2])):
            return True
        if xx!=binsize:
            if xx % step == abs(head[1]-tail[1])%step:
                return True
    return False
#[s[2][3:],int(s[3]),file_order,mismatch,0] [chr,startpos,fileorder,mismatch,0]
#chromosome,startpos,order,mismatch,sum
def overlap(a,b,step,length_bin):
#Detect 2 kinds of overlapping
# 1. fragments combination overlap: fragment 1 and fragment 2,3,4.. can not appear in two different combinations
# 2. combinations can not overlap in reference genome.
    ordera = a[2]
    orderb = b[2]
    if (orderb<ordera):
        k=a
        a=b
        b=k
    ordera = a[2]
    orderb = b[2]
    sa = a[1]
    sb = b[1]
    suma = a[4]
    sumb = b[4]
    if (a[0]==b[0]) and ((sa<sb and sa+suma*step+length_bin>sb) or (sb<sa and sb+sumb*step+length_bin>sa)): 
        return True
    if orderb-ordera-suma<=(length_bin/step)-1: return True
    return False

def RemoveFastqExtension(name):
    '''
    Tested
    This function is used for remove extensions like .gz .fq .fastq.gz
    '''
    newname = name
    if '/' in newname:
        newname = newname[newname.rfind('/')+1:]
    if newname[-3:].lower()=='.gz':
        newname=newname[:-3]
    if newname[-3:].lower()=='.fq':
        newname=newname[:-3]
    if newname[-6:].lower()=='.fastq':
        newname=newname[:-6]
    return newname

def toolcheck(command):
    '''
    Tested
    '''
    t = subprocess.Popen(command,shell=True,stdout = subprocess.PIPE,stderr = subprocess.PIPE)
    out = t.stdout.read().decode()
    err = t.stdout.read().decode()
    if ' command not found' in out or ' command not found' in err:
        return False
    return True

def samtoolsversion():
    t = subprocess.Popen('samtools --version',shell=True,stdout = subprocess.PIPE,stderr = subprocess.PIPE)
    out = t.stdout.read().decode()
    err = t.stdout.read().decode()
    version = out.split('\n')[0].split()[1].strip()
    version = float(version)
    return version

def readf(filename):
    '''
    Tested
    '''
    with open(filename) as f:
        lines = f.readlines()
    return lines

def union(files):
    '''
    Tested
    '''
    dic={}
    for f in files:
        lines = readf(f)
        for line in lines:
            temp = line.strip().split()
            if temp[3]=='.': continue
            key=temp[0]+'\t'+temp[1]+'\t'+temp[2]
            if key in dic:
                dic[key].append(float(temp[3]))
            else:
                dic[key]=[float(temp[3])]
    del_arr=[]
    for key in dic:
        if len(dic[key])!=len(files):
            del_arr.append(key)
    for k in del_arr:
        del(dic[k])
    return dic
            
def exist(file):
    for f in file:
        if isinstance(f,list):
            for ff in f:
                if not os.path.exists(ff):
                    return ff,False
        else:
            if not os.path.exists(f):
                return f,False
    return '',True


def removeFileIfExist(name):
    if os.path.exists(name):
        print('Removing temporary file: '+name+'  ......')
        os.remove(name)
    else:
        print("Temporary file doesn't exist: "+name+' !')


def bam_file_name_format(s):
    return s.strip().split(',')


def fastq_file_name_format(s):
    #split samples using blank. split double end files using comma(,);
    ans=[]
    for f in s:
        if ',' in f:
            ff=f.strip().split(',')
        else:
            ff = [f]
        ans.append(ff)
    return ans


def update_read(read):
    l = len(read.seq)
    read.cigartuples = [(0,l)]
    read.template_length = l


def split_log_localize(read_num_dist, read_len_dist, name):
    result= ['Number of fragments from one read\tread count\n']
    total_ori_reads = 0
    total_frag = 0
    for i,num in enumerate(read_num_dist):
        if num==0: continue
        read_num = i#+1
        result.append(str(read_num)+'\t'+str(num)+'\n')
        total_ori_reads += num
        total_frag += num * read_num
    result.append('Total rescued original reads: '+str(total_ori_reads)+'\n')
    result.append('Total rescued fragments: '+str(total_frag)+'\n')
    result.append('Read length\tread count\n')
    total_bp = 0
    for i,num in enumerate(read_len_dist):
        if num==0: continue
        read_len = i#+1
        result.append(str(read_len)+'\t'+str(num)+'\n')
        total_bp += read_len*num
    result.append('Total base pair: '+str(total_bp)+'\n')
    with open(name,'w') as f:
        f.writelines(result)


if __name__=="__main__":
    #dic=union(['BED_FILE/head_combine.bam.G.bed.short.bed','BED_FILE/head_combine.bam.G.bed.short.bed'])
    #print(dic)
    #print(RemoveFastqExtension('SRR1248444_2.fastq.gz'))
    #files=[]
    #for i in range(15):
    #    files.append('../example/scWGBS/BED_FILE/'+str(i)+'.bed')
    #data=[]
    #dic=union(files)
    #columns = ['chrom','start','end']
    #print(methdic)
    #methdata=list(map(lambda x:x.split()+dic[x],dic))
    #columns.extend(list(map(lambda x:'F'+str(x),list(range(1,len(methdata[0])-2)))))
    #columns.extend(['MII','MII','MII','MII','MII','2iESC','2iESC','2iESC','2iESC','2iESC','Ser_ESC','Ser_ESC','Ser_ESC','Ser_ESC','Ser_ESC'])
    #print(methdata)
    #print(columns)
    #from pandas import DataFrame
    #from bsplot import *
    #df = DataFrame(methdata,columns=columns)
    #point_cluster(df,'point_clutser.png')
    split_log_localize([1,2,3,4,4,5,], [1,2,3,4,5,6,7], 'aa.temp')



import numpy as np
import math
import random
import sys

#==============================Variable setting===================================
readsname = sys.argv[1]
readsnum= int(sys.argv[2])
conversion_ratio=1
methylation_ratio=0.75
read_length=150
'''
@Variables:
chrom_set: dictionary, {chromsome_name:chromsome_genome}
chrom_len: int[], chromsome length ordered by chromsome name in fastq file
chrom_name: string[], chromsome names
genome_len: int, length of whole genome
methy_dict: dictionary, {'chr1_10469':0.1(methylation ratio)}
'''
#================================================================================

def bed_reader(filename):
    '''
        read bed file as fixed methylation ratio
    '''
    dict = {}
    with open(filename) as f:
        for line in f:
            line_content = line.strip().split()
            dict[line_content[0]+'_'+line_content[1]] = float(line_content[3])
    return dict

def genome_loader(filename):
    chr = ''
    chrom_set = {}
    seq = ''
    genome_len = 0
    chrom_len = []
    chrom_name = []
    with open(filename) as f:
        for line in f:
            line_content = line.strip()
            if line_content[0]=='>':
                if chr!='':
                    chrom_set[chr] = seq
                    chrom_len.append(len(seq))
                    genome_len += chrom_len[-1]
                    seq=''
                chr = line_content[1:]
                chrom_name.append(chr)
                continue
            seq += line_content
        chrom_set[chr] = seq
        chrom_len.append(len(seq))
        genome_len += chrom_len[-1]
    return chrom_set, chrom_len, chrom_name, genome_len


def fake_read(length):

    base = ['A','T','C','G']
    seq = ''
    for i in range(length):
        pos = random.randint(0,3)
        seq += base[pos]
    return seq


def reverse(read):
    '''
    CG......
    ......CG
    CG......
    '''
    dic={'A':'T','T':'A','C':'G','G':'C','N':'N'}
    r=''
    for rr in read:
        r=dic[rr.upper()]+r
    return r


def bisulfite(chrom, start, read, fb):
    '''
    return bisulfited seq.
    Ignore the reads if get ''
    '''
    r=''
    l = len(read)
    for i in range(1,l-1): #waste 2 bp
        base=read[i].upper()
        if base=='C':
            c = random.random()
            if c<conversion_ratio:
                if read[i+1].upper()=='G':
                    pos = chrom+'_'+str(start+i)
                    if pos not in methy_dict:
                        methy_value = methylation_ratio
                    else:
                        methy_value = methy_dict[pos]
                    m = random.random()
                    if m>methy_value:
                        base='T'
                    fb.append(pos+'\t'+str(int(m<=methy_value))+'\tCpG\n')
                else:
                    base='T'
        r=r+base
    return r



'''
@Variables:
chrom_set: dictionary, {chromsome_name:chromsome_genome}
chrom_len: int[], chromsome length ordered by chromsome name in fastq file
chrom_name: string[], chromsome names
genome_len: int, length of whole genome
methy_dict: dictionary, {'chr1_10469':0.1(methylation ratio)}
'''

def random_head_tail():
    finalbed=[]
    fake_length = 40
    real_read_length = read_length - fake_length
    reads_count = 0
    while True:
        pos = random.randint(0,genome_len)
        chr=0
        real_read_length = read_length - fake_length
        while pos>=chrom_len[chr]:
            pos-=chrom_len[chr]
            chr+=1
        if pos==0:
            pos+=1
        start=pos-1
        end=pos+real_read_length+1
        if end>chrom_len[chr]: continue
        if pos<1:continue
        # if pos>chrom_len[chr]:continue
        read=chrom_set[chrom_name[chr]][start:end]
        if 'N' in read:
            continue
        r=read
        a1=random.random()
        a2=random.random()
        if a1>0.5:
            r=reverse(read)#Get reads from +/- strand
        r = bisulfite(chrom_name[chr],start,r,finalbed)
        if r=='': continue
        if a2>0.5:
            r=reverse(r)#PCR +/-
        fake_marker=''
        if fake_length>0:
            head_fake_length = random.randint(0,fake_length)
            tail_fake_length = fake_length - head_fake_length
            head = fake_read(head_fake_length)
            tail = fake_read(tail_fake_length)
            r = head + r + tail
            fake_marker = '_'+str(head_fake_length)+'_'+str(tail_fake_length)
        quality = 'E'*read_length
        print('@'+str(reads_count)+'_'+readsname)
        print(r)
        print('+')
        print(quality)
        finalbed.append(chrom_name[chr]+'\t'+str(start+1)+'\t'+str(end-1)+'\t'+str(reads_count)+'_'+readsname+'\t'+str(head_fake_length)+'\t'+str(tail_fake_length)+'\t'+str(a1)+'\t'+str(a2)+'\tread\n')
        reads_count += 1
        if reads_count == readsnum:
            break
    with open(readsname+'_simulation.bed','w') as f:
        f.writelines(finalbed)

chrom_set, chrom_len, chrom_name, genome_len = genome_loader('hg38.fa')
methy_dict = bed_reader('./hESC.bed')#'./hESC.bed')

if __name__=="__main__":
    '''
    print(chrom_set['chr1'][10468:10498])
    print(genome_len)
    print(chrom_len)
    print(chrom_name)
    print(methy_dict['chr1_10468'])
    '''
    random_head_tail()
    



from datetime import datetime
from .mapreduce import *
import multiprocessing
#from Comb_fastq import combine
import pysam
from .utils import *
import sys
import gzip
import os
from .pairend import *

max_file_length = 500 # 24

_index = []
for i in range(max_file_length):
    _index.append(str(i))

reads_cut_buff = {}

def cut(step,length_bin,link,i):
    start = step*i
    end = length_bin + start
    if end > len(link):
        end = len(link)
    if start>=end or end-start<=length_bin-step:
        return False,''
    return True,link[start:end]

def cut_serial(step, length_bin, read_len, max_file_length):
    #read_len = len(link)
    cutted_list = []
    for i in range(max_file_length):
        start = step*i
        end = length_bin + start
        if end > read_len:
            end = read_len
        if start>=end or end-start<=length_bin-step:
            break

        cutted_list.append([start,end])
    return cutted_list

class file_writer:
    
    def __init__(self,step,length_bin,max_file_length,outputname):
        self.step = step
        self.length_bin = length_bin
        self.max_file_length = max_file_length
        self.outputname = outputname
        self.unmapped_file = outputname + '.unmapped.fastq'
        self.left_reads = []
        self.gz = False
        removeFileIfExist(self.unmapped_file)
    
    def gzopen(self):
        self.gz = True
        self.unmapped_file = self.unmapped_file + '.gz'
        removeFileIfExist(self.unmapped_file)

    def file_open(self):
        if self.gz:
            self.f = gzip.open(self.unmapped_file,'w')
        else:
            self.f = open(self.unmapped_file,'w')
        
    def file_close(self):

        self.f.close()

    def writefiles(self,UnmappedReads):
        #unmapped_file = outputname + '.unmapped.fastq.gz'
        batch_size = 50000
        filecontent = []
        for readsname in UnmappedReads:
            link = UnmappedReads[readsname]
            read_length = len(link[0])
            if read_length in reads_cut_buff:
                start_end_list = reads_cut_buff[read_length]
            else:
                start_end_list = cut_serial(self.step,self.length_bin, read_length,self.max_file_length)
                reads_cut_buff[read_length] = start_end_list
            for ind, start_end in zip(_index,start_end_list):
                start,end = start_end
                fc = '@'+readsname+'&'+ind+'\n'+link[0][start:end]+'\n+\n'+link[1][start:end]+'\n'
                #if self.gz: fc = fc.encode()
                filecontent.append(fc)
            if batch_size-len(filecontent)<=24:
                while len(filecontent)<batch_size:
                    t = '@FAKE'+readsname+'\nATCGATCGATCGATCGATCGATCGATCGATCGATCGATCG\n+\n****************************************\n'
                    if self.gz: t = t.encode()
                    filecontent.append(t)
                content = ''.join(filecontent)
                if self.gz: content = content.encode()
                self.f.write(content)
                filecontent = []
        content = ''.join(filecontent)
        if self.gz: content = content.encode()
        self.f.write(content)
        #with gzip.open(unmapped_file,'a') as f:
        #    f.writelines(filecontent)
        #return unmapped_file


def clip_process(inputfileinfo,param,given_bam_file,given_label):
    print('[ '+str(datetime.now())+' ]\tBegin the alignment process with clipping.')
    input_file_num = len(inputfileinfo)
    if input_file_num<=0 or input_file_num>2:
        print("Parameter error in "+l)
        sys.exit()
    refpath = param['ref']
    filter = param['filter_len']
    threads = param['threads']
    step = param['step']
    length_bin = param['window']
    bamu = False
    if 'bamu' in param and param['bamu']:
        bamu=True

    purename = inputfileinfo[0][inputfileinfo[0].rfind('/')+1:]
    
    if not given_bam_file:
        outputname = RemoveFastqExtension(purename)
    else:
        purename_bam = given_bam_file[given_bam_file.rfind('/')+1:]
        outputname = RemoveFastqExtension(purename)+'_'+purename_bam[:-4]
        if param['moabs']:
            outputname = purename_bam[:-4]


    # Part_Fastq_Filename=[]
    #unmapped_file = outputname + '.unmapped.fastq.gz'
    #removeFileIfExist(unmapped_file)
    
    # for i in range(24):
    #     name = outputname+'.part'+str(i+1)+'.fastq'
    #     if os.path.exists(name):
    #         os.system('rm '+name)
    if not given_bam_file:
        output_prefix = outputname
    else:
        output_prefix = given_bam_file[:given_bam_file.rfind('.')]
    
    if param['moabs']:
        store_file_prefix = ''
    else:
        store_file_prefix = 'BAM_FILE/'


    if not given_bam_file:
        originallogname = store_file_prefix+outputname+'_originallog.record'
    else:
        originallogname = ''

    phred=33
    print('[ '+str(datetime.now())+' ]\tBegin the 1st round full reads mapping.')
    if not given_bam_file:
        if input_file_num==2 :
            command='bsmap -a '+inputfileinfo[0]+' -b '+inputfileinfo[1]+' -z '+str(phred)+' -d '+refpath+' -o '+outputname+'.bam -U -S 123 -n 1 -r 0 -u -p ' + threads + ' 1>>LiBis_log 2>>'+originallogname
        else:
            command='bsmap -a '+inputfileinfo[0]+' -z '+str(phred)+' -d '+refpath+'  -o '+outputname+'.bam -S 123 -n 1 -r 0 -U -u -p ' + threads + ' 1>>LiBis_log 2>>'+originallogname
        First_try = Pshell(command)
        First_try.process()
    # --------------------------------------------Multiple mapping for multiple matched reads in pairend mapping.----------------------------------
    print('[ '+str(datetime.now())+' ]\tFiltering and clipping unmapped reads.')
    if 'multiple' in param and param['multiple']:
        ori_bam = outputname+'.bam'
        if given_bam_file:
            ori_bam = given_bam_file
            # pairend_remap(bam_filename,fastq_filename,output_name,refpath, multihits_num=10)
            remove_list = []
            multiple_recovered_bam = pairend_remap(ori_bam, inputfileinfo, outputname, refpath, output_prefix, multihits_num=10)
            remove_list.append(multiple_recovered_bam)
            remove_list.append(multiple_recovered_bam+'.bai')
            multiple_recovered_bam = bam_merge(ori_bam, multiple_recovered_bam)
            for r in remove_list:
                removeFileIfExist(r)
    file_w = file_writer(step,length_bin,max_file_length,outputname)
    if 'gz' in param and param['gz']:
        file_w.gzopen()
    file_w.file_open()

    if given_bam_file and not bamu:
        set_sam = {}
        set_sam_pair = set()
        bam_file_name = given_bam_file
        bam_file = pysam.AlignmentFile(bam_file_name,'rb')
        for line in bam_file:
            qname = line.query_name
            next_name = line.next_reference_name
            next_pos = line.template_length
            if next_name!=None:
                if next_pos>0:
                    set_sam_pair.add(qname)
            else:
                if input_file_num==2:
                    m1 = (line.flag & 64)
                    m2 = (line.flag & 128)
                    if m1>0: mate = 1
                    elif m2>0: mate = 2
                    else: mate = 0
                else:
                    mate = 0
                set_sam[qname]=mate
        
        UnmappedReads = {}
        o=0
        step = param['step']
        length_bin = param['window']
    
        for filename in inputfileinfo:
            o+=1
            gzmark=False
            if filename.endswith('.gz'):
                f = gzip.open(filename)
                gzmark=True
            else:
                f = open(filename)
    
    
            if f:
                while 1:
                    if gzmark:
                        line1 = f.readline().decode()
                    else:
                        line1 = f.readline()
                    if not line1:
                        break
                    if gzmark:
                        line2 = f.readline().decode().strip()
                        line3 = f.readline().decode()
                        line4 = f.readline().decode().strip()
                    else:
                        line2 = f.readline().strip()
                        line3 = f.readline()
                        line4 = f.readline().strip()
                    line1 = line1.strip().split()
                    read_name = line1[0][1:]
                    pair_file_rank = o
                    if '/' in read_name:
                        split_pos=0
                        if '/' in read_name:
                            split_pos = read_name.rfind('/')
                        read_name = read_name[:split_pos]
    
                    if read_name in set_sam_pair: continue
                    mapped_mate = set_sam.get(read_name,-1)
                    if mapped_mate==0 or mapped_mate==pair_file_rank: continue
    
                    if input_file_num>1: read_name+='_'+str(pair_file_rank)
                    # Maybe the mate search method is buggy. Cuz there are different structures of reads name generated by different sequencing machine.
                    # Fixed in issue 1 in github.
                    UnmappedReads[read_name]=[line2,line4]
                    # Just adjust the Unmapped output size to 50000 and discard all small reads over 50000
                    if len(UnmappedReads)>3000000:
                    #if len(UnmappedReads)>50000:
                        file_w.writefiles(UnmappedReads)
                        UnmappedReads={}
                        
    
        if len(UnmappedReads)>0:
            file_w.writefiles(UnmappedReads)
        f.close()
        del UnmappedReads
        # sys.exit()    
        # We've got the splited fastq file, filename is stored in Part_Fastq_Filename
        # for i in range(len(Part_Fastq_Filename)):
    else:
        bam_file_name = outputname+'.bam'
        if given_bam_file:
            bam_file_name = given_bam_file
        bam_file = pysam.AlignmentFile(bam_file_name,'rb')
        UnmappedReads = {}
        for line in bam_file:
            if (line.flag & 4 != 4): continue
            strand_info=''
            if line.flag & 64 == 64:
                strand_info='_1'
            if line.flag & 128 == 128:
                strand_info='_2'
            qname = line.query_name + strand_info
            seq = line.seq
            qua = line.qual
            UnmappedReads[qname] = [seq, qua]
            if len(UnmappedReads)>3000000:
                file_w.writefiles(UnmappedReads)
                UnmappedReads={}
        if len(UnmappedReads)>0:
            file_w.writefiles(UnmappedReads)
            del UnmappedReads
        bam_file.close()
    unmapped_file = file_w.unmapped_file
    file_w.file_close() 
    print('[ '+str(datetime.now())+' ]\tBegin the 2st round mapping for clipped fragments.')
    command = 'bsmap -a '+unmapped_file+' -d '+refpath+'  -o '+unmapped_file[:unmapped_file.find('fastq')-1]+'.bam -n 1 -r 0 -U -R -p ' + threads + ' 1>>LiBis_log 2>>'+store_file_prefix+unmapped_file[:unmapped_file.find('fastq')-1]+'_log.txt'
    Batch_try = Pshell(command)
    Batch_try.process()
    print('[ '+str(datetime.now())+' ]\tRecombining uniquely mapped fragments...')
    # run bsmap and get bam files named as Part_Fastq_Filename[i].bam
    # import combine to generate the finalfastq

    args={'step':param['step'],
          'binsize':param['window'],
          'filter':filter,
          'outputname':outputname,
          'originalfile':inputfileinfo,
          #'mapfilenumber':10,
          #'report_clip':param['report_clip']
          #'finish':1
         }

    reads_num_dist,reads_len_dist = unsortedCombine(unmapped_file[:unmapped_file.find('fastq')-1],args)
    #mapreduce_names = reads_map(unmapped_file[:-3],args)
    #reads_reduce(mapreduce_names,args)
    splitlogname = store_file_prefix+outputname+'_split_log.record'
    split_log_localize(reads_num_dist,reads_len_dist,splitlogname)
    #command = 'bsmap -a '+outputname+'_finalfastq.fastq.gz -d '+refpath+' -o '+outputname+'_split.bam -S 123 -n 1 -r 0 -p ' + threads + ' 1>>LiBis_log 2>> '+splitlogname
    #Bam = Pshell(command)
    #Bam.process()
    splitfilename = outputname+'_split.bam'
    #------------------------------------------------Pairend filter for LiBis rescued reads----------------------------------------------------------
    if 'pairend' in param and param['pairend']:
        ori_bam = outputname+'.bam'
        if given_bam_file:
            ori_bam = given_bam_file
        PEfilter_bam = libis_filter(ori_bam, splitfilename)
        command = 'mv '+PEfilter_bam+' '+splitfilename
        renamer = Pshell(command)
        renamer.process()
        #libis_filter(ori_bam, split_bam)
    #------------------------------------------------------------------------------------------------------------------------------------------------
    
    print('[ '+str(datetime.now())+' ]\tBAM file sorting...')
    #command='samtools sort -f -@ 6 '+splitfilename+' '+splitfilename+'.sorted.bam'
    #filter = Pshell(command)
    #filter.process()
    pysam.sort('-@',threads,'-o',splitfilename+'.sorted.bam','-O','BAM',splitfilename)
    #pysam.sort('-@',6,'-n',splitfilename,splitfilename+'.sorted')
    if not given_bam_file:
        #command='samtools sort -f -@ 6 '+outputname+'.bam'+' '+outputname+'.sort.bam'
        command = 'samtools view -b -F 4 -@ 6 '+outputname+'.bam > '+outputname+'.filter.bam'
        filter=Pshell(command)
        filter.process()
        pysam.sort('-@',threads,'-o',outputname+'.sorted.bam','-O','BAM',outputname+'.filter.bam')
        #pysam.sort('-@',6,'-n',outputname+'filter.bam',outputname+'.sorted')
        removeFileIfExist(outputname+'.filter.bam')
        removeFileIfExist(outputname+'.bam')
    else:
        if param['bamu']:
            command = 'samtools view -b -F 4 -@ 6 '+given_bam_file+' > '+outputname+'.filter.bam'
            filter=Pshell(command)
            filter.process()
            pysam.sort('-@',threads,'-o',outputname+'.sorted.bam','-O','BAM',outputname+'.filter.bam')
            #pysam.sort('-@',6,'-n',outputname+'filter.bam',outputname+'.sorted')
        else:
            pysam.sort('-@',threads,'-o',outputname+'.sorted.bam','-O','BAM',given_bam_file)
            #pysam.sort('-@',6,'-n',given_bam_file,outputname+'.sorted')
        removeFileIfExist(outputname+'.filter.bam')
        #command='samtools sort -f -@ 6 '+given_bam_file+' '+outputname+'.sort.bam'
    #filter.change(command)
    #filter.process()
    #command='mv '+outputname+'.sort.bam '+outputname+'.bam'
    #command = 'samtools view -b -F 4 -@ 6 '+outputname+'.sort.bam > '+outputname+'.bam' 
    #filter.change(command)
    #filter.process()
    #command='mv '+splitfilename+'.sorted.bam '+splitfilename
    #filter.change(command)
    #filter.process()
    if os.path.exists(splitfilename+'.sorted.bam'):
        os.rename(splitfilename+'.sorted.bam', splitfilename)
    if os.path.exists(outputname+'.sorted.bam'):
        os.rename(outputname+'.sorted.bam',outputname+'.bam')

    
    if not param["moabs"]:
        filter = Pshell("")
        #if given_bam_file:
        #    command='cp '+given_bam_file+' '+outputname+'.bam'
        #    filter.change(command)
        #    filter.process()
        m=Pshell('samtools merge -f -@ 6 '+store_file_prefix+outputname+'_combine.bam '+outputname+'.bam '+splitfilename)
        m.process()
        command='mv '+outputname+'.bam BAM_FILE/'
        filter.change(command)
        filter.process()
        command='mv '+splitfilename+' BAM_FILE/'
        filter.change(command)
        filter.process()
        pysam.index(store_file_prefix+outputname+'_combine.bam')
        pysam.index('BAM_FILE/'+outputname+'.bam')
        pysam.index('BAM_FILE/'+splitfilename)
    else:
        pysam.index(outputname+'.bam')
        pysam.index(splitfilename)
    print('[ '+str(datetime.now())+' ]\tFINISH!')
    return 'BAM_FILE/'+outputname+'_combine.bam',originallogname,splitlogname,[unmapped_file,outputname+'_finalfastq.fastq.gz',outputname+'.sam']
    #print("Merge done!")#\nCreated final bam file called "+outputname+'_combine.bam')

def clipmode(name,param, given_bam_file,given_label):
    '''
    When we get the mapping result, we should report 
    mapping ratio, mapped reads number, length distribution,
    original mapping ratio, original mapped reads number,
    new generated splitted reads number, new generated splitted reads length
    '''
    print('[ '+str(datetime.now())+' ]\tTemprary files cleaning...')
    newn,originallog,splitlog,cleanname=clip_process(name,param, given_bam_file,given_label)

    if (not 'cleanmode' in param) or param['cleanmode']:
        #Set a clean mode and full mode for clipping mode
        cleanupmess(cleanname)

    return newn,[originallog,splitlog]


def cleanupmess(name):
    unmapped_file,n1,n2 = name
    outputname = n1[:n1.rfind('_')]
    #removeFileIfExist(n1)
    #removeFileIfExist(n2)
    removeFileIfExist(outputname+'.sort.bam')
    removeFileIfExist(outputname+'.bam.bai')
    #removeFileIfExist(outputname+'_split.bam.bai')
    removeFileIfExist(unmapped_file)
    unmapped_name = unmapped_file[:unmapped_file.rfind('fastq')-1]
    removeFileIfExist(unmapped_name+'.bam')
    #removeFileIfExist(unmapped_file[:-3]+'.sam')
    removeFileIfExist(unmapped_name+'.bam.bai')
    #for i in range(10):
    #    removeFileIfExist(outputname+'_'+str(i)+'.mapreduce')
    


if __name__=="__main__":
    
    cleanupmess(['6P1_notrim_val_1_val_1_test.unmapped.fastq','6P1_notrim_val_1_val_1_test_finalfastq.fastq','6P1_notrim_val_1_val_1_test.sam'])
    #with open("config.txt") as f:
    #    lines = f.readlines()
    #import multiprocessing
    #pool = multiprocessing.Pool(processes=2)
    #for l in lines:
    #    #pool.apply_async(clip_process,(l,))
    #    clip_process(l.strip().split(),{'ref':'/data/dsun/ref/humanigenome/hg19.fa','step':5,'window':30,'cleanmode':True}) #pass file name to clip_process
    #pool.close()
    #pool.join()
    #p = {'ref':'/data/dsun/ref/humanigenome/hg19.fa','step':5,'window':30,'cleanmode':True}
    #clipmode(['Trim/head.fq'],p)



'''
Tested
'''
import matplotlib
matplotlib.use('Agg')
import numpy as np 
import matplotlib.pyplot as plt 
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
import os
from urllib import request
from copy import deepcopy
import matplotlib.cm as cm

def point_cluster(data,outputname,method='PCA'):
    #data: DataFrame
    #contains labelname column, samplename column and data
    d = deepcopy(data)
    #print(di)
    d.sort_values(['chrom','start'])

    windowdata = d.values[:,3:].T
    position = d.values[:,:3]
    label = d.columns[3:]
    #Every sample in a row
    #print(windowdata)
    dim=2
    colors = cm.rainbow(np.linspace(0,1,len(label)))
    if method == 'PCA':
        pca = PCA(n_components=dim)
        x_tr = pca.fit_transform(windowdata)
    if method == 'TSNE':
        tsne = TSNE(n_components=dim)
        x_tr = tsne.fit_transform(windowdata)
    fig,ax = plt.subplots(figsize=(7,7))
    xmin = np.min(x_tr[:,0])
    xmax = np.max(x_tr[:,0])
    ymin = np.min(x_tr[:,1])
    ymax = np.max(x_tr[:,1])
    sample_size = x_tr.shape[0]
    plt.xlim(xmin-0.2*np.abs(xmin),xmax+0.2*np.abs(xmax))
    plt.ylim(ymin-0.2*np.abs(ymin),ymax+0.2*np.abs(ymax))
    markers=['o', '^','v','<','>','1','2', '3','4','8','s','P','p', '*','H','h','x','X','D']
    label_u = np.unique(label)
    for i in range(len(label_u)):
        cc = colors[i]
        l = label_u[i]
        pos = np.where(label==l)
        ma = markers[i]
        plt.scatter(x_tr[pos,0],x_tr[pos,1],c=cc,alpha=0.8,s=50,marker=ma,label=l)

    if method=='PCA':
        plt.xlabel('PC1',fontsize=13)
        plt.ylabel('PC2',fontsize=13)
    if method=='TSNE':
        plt.xlabel('TSNE1',fontsize=13)
        plt.ylabel('TSNE2',fontsize=13)
    plt.legend(loc='best',fontsize=13)
    plt.savefig(outputname)

def heatmap(data,outputname):
    from seaborn import clustermap 
    d = deepcopy(data)
    #print(d)
    d=d.drop(['chrom','start','end'],axis=1)
    sns_plot=clustermap(d)
    sns_plot.ax_row_dendrogram.set_visible(False)
    sns_plot.savefig(outputname)


if __name__=='__main__':
    with open('BED_FILE/head_combine.bam.G.bed.short.bed') as f:
        lines = f.readlines()
    d=[]
    for line in lines:
        temp = line.strip().split()
        temp[-1]=float(temp[-1])
        d.append(temp)
    import random
    for i in range(5):
        for dd in d:
            dd.append(random.random())
    import pandas as pd
    d = pd.DataFrame(d,columns=['chrom','start','end','L1','L2','L3','L4','L4','L4'])
    heatmap(d,'a.pdf')


    


    



    



from .utils import *

def libisSplitExtractor(filename):
    lines = readf(filename)
    rescued_reads, rescued_frags = 0,0
    for line in lines:
        if 'Total rescued original reads:' in line:
            c = line.strip().split()
            rescued_reads = int(c[-1])
        if 'Total rescued fragments:' in line:
            c = line.strip().split()
            rescued_frags = int(c[-1])
    return 0, rescued_reads, rescued_frags

def BsmapOutputExtractor(filename):
    #aligned=[]
    #unique=[]
    #total=[]
    lines = readf(filename)
    countinputfile=sum(list(map(lambda x: 'Input read file' in x,lines)))
    total=0
    unique=0
    aligned=0
    if countinputfile==2:
        for line in lines:
            if "total read pairs:" in line:
                a = line.strip().split()
                total=int(line[line.rfind('total read pairs:')+len('total read pairs:'):].split()[0])*2                    
                continue
            if "aligned pairs:" in line:
                a = line.strip().split()
                aligned = aligned+int(a[2])*2
                unique = unique + int(a[6])*2
                continue
            if "unpaired read #1:" in line or "unpaired read #2:" in line:
                a = line.strip().split()
                aligned = aligned+int(a[3])
                unique = unique+int(a[7])
    else:
        for line in lines:
            if "total reads:" in line:
                a = line.strip().split()
                # print(a)
                total=int(line[line.rfind('total reads:')+len('total reads:'):].split()[0])
                continue
            if "aligned reads:" in line:
                a = line.strip().split()
                aligned = aligned+int(a[2])
                unique = unique+int(a[6])
    return [total,aligned,unique]


def McallOutputExtractor(filename):
    lines = readf(filename)


def TrimOutputExtractor(filename):
    '''
    Tested
    '''
    ratio=[]
    for nn in filename:
        temp=[]
        for name in nn: 
            lines = readf(name)
            #Total written (filtered):  1,085,317,679 bp (94.2%)
            for line in lines:
                temp = line.strip().split()
                if len(temp)<5: continue
                r=None
                if temp[0]=='Total' and temp[1]=='written' and temp[2]=='(filtered):':
                    r=float(temp[5].strip()[1:-2])
                    break
                temp.append(r)
            ratio.append(temp)
    return ratio

'''
Result from BsmapOutputExtractor:
    total reads, mapped reads, uniquely mapped reads
Result from BsmapResult:
    [[total reads,mapped reads,uniquely mapped reads, clipped reads, unique clipped reads,
    all mapped reads, all uniquely mapped reads, mapping ratio, uniquely mapping ratio],...]
'''
def BsmapResult(filenames,clip,offered_bamfile):
    if clip:
        '''
        We will get two record files in this mode
        '''
        result=[]
        # print(filenames)
        # print(offered_bamfile)
        for (name, processed_bam) in zip(filenames,offered_bamfile):
            ori,spl = name
            ori_info = BsmapOutputExtractor(ori) if processed_bam=='' else [1,1,1]
            spl_info = libisSplitExtractor(spl)#BsmapOutputExtractor(spl)
            total_reads,mapped_reads,uniquely_mapped_reads=ori_info
            _,clipped_reads,clipped_frags=spl_info
            result.append([total_reads,
                mapped_reads,
                uniquely_mapped_reads,
                clipped_reads,
                clipped_frags,
                mapped_reads+clipped_reads,
                uniquely_mapped_reads+clipped_reads,
                (mapped_reads+clipped_reads)/float(total_reads),
                (uniquely_mapped_reads+clipped_reads)/float(total_reads)
            ])
            #print(result)
            if processed_bam!='':
                for i in [0,1,4,5,6,7,8]:
                    result[-1][i]='Not Available'
    else:
        '''
        Only one record file need to be processed.
        '''
        if processed_bam!='':
            result.append(['Not Available']*9)
            return result
        result=[]
        for name in filenames:
            ori = name
            ori_info = BsmapOutputExtractor(ori)
            #print(ori_info)
            total_reads,mapped_reads,uniquely_mapped_reads=ori_info
            _,clipped_reads,clipped_frags=[0,0,0]
            result.append([total_reads,mapped_reads,uniquely_mapped_reads,clipped_reads,
                clipped_frags,
                mapped_reads+clipped_reads,uniquely_mapped_reads+clipped_reads,
                (mapped_reads+clipped_reads)/float(total_reads),
                (uniquely_mapped_reads+clipped_reads)/float(total_reads)])
    return result

if __name__=="__main__":
    #r=BsmapResult(['BAM_FILE/temp'],False)
    r=libisSplitExtractor('/data/jiali/map.log/Tis183.R1.am.trim_split_log.record')
    print(r)



from .moabs import *
from .fastqc import Fastqc
from .trim import Trim
from .utils import *
from .NumExtractor import *
from .bedtools import Bedtools
from pandas import DataFrame
from .bsplot import *
import os
from datetime import datetime

def computeQC(name,fastqc,param={}):
    '''
    Computing function for qc. Return two arrays: The first one contains filenames for next step computing.
    The second one contains 
    '''
    print('[ '+str(datetime.now())+' ]\tRunning quality control for samples...')
    fn=[]
    filename=''
    for nn in name:
        temp=[]
        for n in nn:
            filename = filename+n+' '
            newname = RemoveFastqExtension(n)
            temp.append(newname+'_fastqc.html')
        fn.append(temp)
    fastqc.run(filename)
    print('[ '+str(datetime.now())+' ]\tFinish quality control for samples')
    return name,fn

def computeTrim(name,trim,param={}):
    print('[ '+str(datetime.now())+' ]\tRunning Adaptor trimming for samples...')
    fn=''
    trimmedname=[]
    triminfo=[]
    #print(name)
    for n in name:
        temp=[]
        infotemp=[]
        fn=''
        fileorder=1
        for nn in n:
            newname = RemoveFastqExtension(nn)
            purename = nn[nn.rfind('/')+1:]
            fn = fn+nn+' '
            if len(n)==1:
                temp.append('Trim/'+newname+'_trimmed.fq.gz')
            else:
                temp.append('Trim/'+newname+'_val_'+str(fileorder)+'.fq.gz')
                fileorder+=1
            infotemp.append('Trim/'+purename+'_trimming_report.txt')
        trimmedname.append(temp)
        triminfo.append(infotemp)
        fn.strip()
        if len(n)==1:
            trim.run(fn,1)
        else:
            trim.run(fn,2)
    name=trimmedname
    print('[ '+str(datetime.now())+' ]\tFinish Adaptor trimming for samples')
    return name,triminfo

def computeBsmap(name,bsmap,param):
    #return ['BAM_FILE/6P1_notrim_val_1_val_1_combine.bam','BAM_FILE/6P1_notrim_val_1_val_1_combine.bam'],[['BAM_FILE/6P1_notrim_val_1_val_1_originallog.record','BAM_FILE/6P1_notrim_val_1_val_1_split_log.record'],['BAM_FILE/6P1_notrim_val_1_val_1_originallog.record','BAM_FILE/6P1_notrim_val_1_val_1_split_log.record']]
    print('[ '+str(datetime.now())+' ]\tRunning Alignment for samples...')
    if not exist(name):
        raise "Fastq not found"
    clip = param['clip']
    given_bam_files = param['bamfiles']
    labels = param['label']
    bamname=[]
    bamlogname=[]
    threads = param['threads']
    for n,given_bam,given_label in zip(name,given_bam_files,labels):
        if clip:
            newname,logname = bsmap.clipping(n,param,given_bam,given_label)
        else:
            if given_bam!='':
                logname = ''
                newname = given_bam
            else:
                newname,logname = bsmap.normalmode(n,given_label,threads)
        bamname.append(newname)
        bamlogname.append(logname)
    print('[ '+str(datetime.now())+' ]\tFinish Alignment for samples')
    return bamname,bamlogname

def computeMcall(name,mcall,param):
    print('[ '+str(datetime.now())+' ]\tRunning Methylation Calling for samples...')
    if not exist(name):
        raise "BAM not found"
    bedname=[]
    logname=[]
    # print(name)
    for n in name:
        # print(n)
        #bn="BED_FILE/6P1_notrim_val_1_val_1_combine.bam.G.bed"
        #ln="BED_FILE/6P1_notrim_val_1_val_1_combine.bam_stat.txt"
        bn,ln=mcall.run(n)
        bedname.append(bn)
        logname.append(ln)
    print('[ '+str(datetime.now())+' ]\tFinish Methylation Calling for samples')
    return bedname,logname

def computeProcess(param):
    fastqc,trim,bsmap,mcall,bedtools = Fastqc, Trim, Bsmap, Mcall, Bedtools
    ProcessObject=[fastqc,trim,bsmap,mcall,bedtools]

    if not param['moabs']:
        for obj in ProcessObject:
            status,word=obj.check(param['nocheck'])
        #if not status:
        #    raise Exception(word)
    
    #check all software have been installed successfully.
    for obj in ProcessObject[:-1]:
        obj.setpath('./')
    
    bsmap.setparam(param)
    mcall.setparam(param)

    name = param['name']
    resultfilename={}
    if param['qc']:
        name,qcresult=computeQC(name,fastqc,{}) 
        resultfilename['qc']=qcresult
        
        if os.path.exists('RESULT') and not os.path.exists('RESULT/qc'):
            os.system('mkdir RESULT/qc')
        os.system('cp Fastqc/*.html RESULT/qc')
    
    if param['trim']:
        name,trimresult = computeTrim(name,trim,{})
        resultfilename['trim']=trimresult

    name,bsmapresult = computeBsmap(name,bsmap,param) 
    resultfilename['bsmap'] = bsmapresult
    originalfilename = param['name']
    filelabel = param['label']
    marker=['Filename','Label']
    marker.append('QC')
    if param['qc']:
        qcresult = []
        prepath = "qc/"
        for q in resultfilename['qc']:
            if len(q)==1:
                qcresult.append('<a href="'+prepath+q[0]+'">QC'+'</a>')
            else:
                qcresult.append('<a href="'+prepath+q[0]+'">QC1'+'</a>,'+'<a href="'+prepath+q[1]+'">QC2'+'</a>')
    else:
        qcresult=["/" for i in range(len(originalfilename))]
    marker.append('Trim')
    if param['trim']:
        trimresult =TrimOutputExtractor(resultfilename['trim'])#,param['clip'])
    else:
        trimresult=["/" for i in range(len(originalfilename))]


    marker.extend(['Input reads','mapped reads','uniquely mapped reads','clipped reads', \
            'clipped fragments','all mapped reads','all uniquely mapped reads',\
            'mapping ratio','uniquely mapping ratio'])
    bsmapresult = BsmapResult(resultfilename['bsmap'],param['clip'],param['bamfiles'])#All contents required by the last extend

    sample=0
    datatable=[marker]
    for orin in originalfilename:
        l = filelabel[sample]
        br = bsmapresult[sample]
        temp=[]
        filenum=0
        if len(orin)==1:
            nn=orin[0]
        else:
            nn=orin[0]+','+orin[1]
        temp.extend([nn,l])
        #if param['qc']:
        temp.append(qcresult[filenum])
        #if param['trim']:
        trim_s = trimresult[sample]
        if trim_s=="/" or len(trim_s)==1:
            temp.append(trimresult[sample][0])
        else:
            temp.append(trimresult[sample][0]+','+trimresult[sample][1])
        filenum=filenum+1
        temp.extend(br)
        sample = sample+1
        datatable.append(temp)
        

    datatableprint = list(map(lambda x: ((''.join(list(map(lambda y:str(y)+'\t',x )))).strip()+'\n'),datatable))
    with open('RESULT/datatable.txt','w') as f:
        f.writelines(datatableprint)
    
    if 'mcall' not in param or not param['mcall']:
        return
    
    name,mcallresult = computeMcall(name,mcall,param)
    resultfilename['mcall'] = mcallresult
    meth_cpg_bed_name=name
    '''
    All computing job is done here. Plotting is behind.
    Table:
    Filename, Label, trim ratio(TrimOutputExtractor),input reads,mapped reads, unique mapped reads,
    clipped reads, unique clipped reads, all mapped reads, unique all, mapping ratio, unique mapping ratio
    
    Plots:
    1. Extract trim ratio from trim report files. Generate a bar plot
    2. Extract bsmap mapping ratio 
    3. 
    
    '''
    if 'mcall' not in resultfilename or 'bsmap' not in resultfilename:
        return
    if 'plot' not in param or not param['plot']:
        return
    print('[ '+str(datetime.now())+' ]\tGenerating reports...')

    '''
    Result from BsmapResult:
    [[total reads,mapped reads,uniquely mapped reads, clipped reads, unique clipped reads,
    all mapped reads, all uniquely mapped reads, mapping ratio, uniquely mapping ratio],...]
    '''
        
    '''
    Table generated as RESULT/datatable.txt
    '''
    #print(param)
    if param['genome']!=None and len(filelabel)>1:
        bedtools.setparam(param)
        bedtools.makewindow()
        shortnames=list(map(lambda x:x+'.short.bed',meth_cpg_bed_name))
        intersectnames=bedtools.intersect(meth_cpg_bed_name)
        methdic=union(intersectnames)
        columns = ['chrom','start','end']
        #print(methdic)
        methdata=list(map(lambda x:x.split()+methdic[x],methdic))
        #columns.extend(list(map(lambda x:'F'+str(x),list(range(1,len(methdata[0])-2)))))
        columns.extend(filelabel)
        #print(methdata)
        #print(columns)
        df = DataFrame(methdata,columns=columns)
        point_cluster(df,'RESULT/point_cluster.png')
        heatmap(df,'RESULT/heatmap.png')
        for sn in shortnames:
            removeFileIfExist(sn)
    abspath = os.path.abspath(__file__)
    pos=abspath.find('computeProcess')
    abspath = abspath[:pos]
    os.system('cp '+abspath+'result.html RESULT/')
    os.system('cp -r '+abspath+'jslib RESULT/')
    removeFileIfExist(param['genome']+'_'+str(param['bin'])+'.bed')
    for i in range(len(filelabel)):
        removeFileIfExist('BED_FILE/'+str(i)+'.intersect')
        removeFileIfExist('BED_FILE/'+str(i)+'.bed')

        
    
if __name__=="__main__":
    param={'name':[['6P1_notrim_val_1.fq.gz','6P2_notrim_val_2.fq.gz'],['6P1_notrim_val_1.fq.gz','6P2_notrim_val_2.fq.gz']], 
           'clip':1, 
           'label':['sample1','sample2'], 
            'window':30, 
            'step':5, 
            'process':1, 
            'ref':'/data/dsun/ref/humanigenome/hg19.fa', 
            'qc':1,
            'trim':1,
            'genome':'hg19',
            'bin':1000000,
            'filter_len':40
        }
    computeProcess(param)   
         
        

        



from .utils import *
import os
import gzip
import pysam

def reads_map(unmapped_file,args):
    mapfilenum = args['mapfilenumber']
    step = args['step']

    file_order=0
    mapreduce_file=[]
    rootfile = unmapped_file[:unmapped_file.find('.')]
    dic={}
    for i in range(mapfilenum):
        mapreduce_file.append(rootfile+'_'+str(i)+'.mapreduce')
        if os.path.exists(mapreduce_file[-1]) and (not 'finish' in args):
            os.remove(mapreduce_file[-1])
            #print('Delete '+mapreduce_file[-1])
        dic[i]=[]
    if 'finish' in args:
        return mapreduce_file

    count=0

    with pysam.AlignmentFile(unmapped_file+'.bam','r') as f:
        for line in f:
            #s = line.strip().split('\t')
            mismatch = line.tags[0][1]#int(s[11][s[11].rfind(':')+1:])
            if mismatch>1: continue
            read_length = line.query_length#len(s[9])
            tail_length = ((read_length-1)%step)+1
            refseq = line.tags[1][1][-2-tail_length:-2]#s[12][-2-tail_length:-2]
            readsseq = line.seq[-tail_length:]#s[9][-tail_length:]
            strand = line.tags[2][1]#s[13][-2:]
            mis=0
            for base in range(tail_length):
                if (refseq[base]!=readsseq[base]):
                    if strand[0]=='+':
                        if (refseq[base]=='C' and readsseq[base]=='T'): continue
                    else:
                        if (refseq[base]=='G' and readsseq[base]=='A'): continue
                    mis+=1
            tail_mismatch = mis
            query_name = line.query_name.split('&')
            #s[0] = s[0].strip()
            read_name = query_name[0]
            file_order = int(query_name[1]) 
            hashnum = abs(hash(read_name)) % mapfilenum
            dic[hashnum].append([read_name,line.reference_name,str(line.reference_start),str(file_order),str(mismatch),str(tail_mismatch),str(read_length)])
            count+=1
            if count>10000000:
                for i in range(mapfilenum):
                    arr = dic[i]
                    arr = list(map(lambda x:'\t'.join(x)+'\n',arr))
                    with open(mapreduce_file[i],'a') as ff:
                        ff.writelines(arr)
                    dic[i]=[]
                count=0
        if count>0:
            for i in range(mapfilenum):
                arr = dic[i]
                arr = list(map(lambda x:'\t'.join(x)+'\n',arr))
                with open(mapreduce_file[i],'a') as ff:
                    ff.writelines(arr)
                dic[i]=[]
            count=0
    return mapreduce_file


def reads_combine(filename,args):
    step = args['step']
    length_bin = args['binsize']
    filter = args['filter']
    outputname = args['outputname']
    originalfile = args['originalfile']
    result={}
    with open(filename) as f:
        for line in f:
            arr = line.strip().split()
            name,chr,startpos,fileorder,mismatch,tail_mismatch,read_length = arr
            startpos = int(startpos)
            fileorder = int(fileorder)
            mismatch = int(mismatch)
            tail_mismatch = int(tail_mismatch)
            read_length = int(read_length)
            if name not in result:
                result[name]=[[chr,startpos,fileorder,mismatch,tail_mismatch,read_length]]
            else:
                result[name].append([chr,startpos,fileorder,mismatch,tail_mismatch,read_length])
    print(len(result))
    for name, content_list in result.items():
        content_list = sorted(content_list, key=lambda x:x[2])
        # initialize result_fragment
        chr,startpos,fileorder,mismatch,tail_mismatch,read_length = content_list[0]
        result_fragment = [[chr,startpos,fileorder,mismatch,0]]
        # Fragment combination
        for c in content_list[1:]:
            chr,startpos,fileorder,mismatch,tail_mismatch,read_length = c
            temp = [chr,startpos,fileorder,mismatch,0]
            join_or_not=False
            for reads in result_fragment:
                if reads[3]+tail_mismatch<=1 and readsjoin(reads,temp,step,read_length,length_bin):
                    reads[3]+=tail_mismatch
                    reads[4]=temp[2]-reads[2]
                    join_or_not=True
            if not join_or_not:
                result_fragment.append(temp)
        frag_num = len(result_fragment)
        for i in range(frag_num-1,-1,-1):
            if result_fragment[i][4]<=1 and result_fragment[i][3]>0:
                result_fragment.pop(i)
        frag_num = len(result_fragment)
        del_mark = [0] * frag_num
        for i in range(frag_num):
            for j in range(i+1,frag_num):
                if overlap(result_fragment[i],result_fragment[j],step,length_bin):
                    sss = result_fragment[i][4] - result_fragment[j][4]
                    if sss>0: del_mark[j]=1
                    elif sss<0: del_mark[i]=1
                    else:
                        mis = result_fragment[i][3] - result_fragment[j][3]
                        if mis>0: del_mark[i]=1
                        else: del_mark[j]=1
        for i in range(frag_num-1,-1,-1):
            if del_mark[i]==1:
                result_fragment.pop(i)
        result[name] = result_fragment
    return result


    

def reads_reduce(mapreduce_file,args):
    step = args['step']
    length_bin = args['binsize']
    filter = args['filter']
    outputname = args['outputname']
    originalfile = args['originalfile']
    mapfilenum = args['mapfilenumber']
    report_clip = args['report_clip']
    if report_clip==None:
        report_clip = False
    removeFileIfExist(outputname+'_finalfastq.fastq.gz')
    if report_clip:
        removeFileIfExist(outputname+'_finalfastq_clipped_head.fastq.gz')
        removeFileIfExist(outputname+'_finalfastq_clipped_tail.fastq.gz')
    totalresult=[{},{}]
    for i in range(mapfilenum):
        result=reads_combine(mapreduce_file[i],args)
        
        #print(len(result))

        pos_mark = [{},{}]
        for name, reads_list in result.items():#nameset:
            #readinfo = nameset[name]
            if len(reads_list)==0: continue
            pos = 0
            
            if name[-2:]=='_1' or name[-2:]=='_2':
                if name[-2:]=='_2':
                    pos = 1
                name = name[:-2]
            for read in reads_list:#readinfo:
                order, sum = read[2], read[4]
                start = (order)*step
                end = start + step*sum + length_bin
                if end-start<filter: 
                    continue
                if name in pos_mark[pos]:
                    pos_mark[pos][name].append([start,end])
                else:
                    pos_mark[pos][name]=[[start,end]]
 


        totalresult[0].update(pos_mark[0])
        totalresult[1].update(pos_mark[1])
        content_number = len(totalresult[0])+len(totalresult[1])
        if content_number>5000000 or (i==mapfilenum-1 and content_number>0):
            GetFastqList(totalresult,step,length_bin,filter,outputname,originalfile,report_clip)
            totalresult={}


def GetFastqList(joined_reads,step,length_bin,filter,outputname,originalfile,report_clip):
    pos_mark = joined_reads
    fileorder=0
    result=[]
    result_begin=[]
    result_end=[]
    for file in originalfile:
        gzmark=False
        if file.endswith('gz'):
            gzmark=True
        if not gzmark:
            f = open(file)
        else:
            f = gzip.open(file)

        while True:
            name = f.readline()
            if not name:
                break
            reads = f.readline()
            _ = f.readline()
            quality = f.readline()
            if gzmark:
                name = name.decode()
                reads = reads.decode()
                quality = quality.decode()

            reads = reads.strip()
            quality = quality.strip()
            fqname = name.strip().split()[0][1:]
            if '/' in fqname[-5:]: # or '.' in fqname:
                split_pos = fqname.rfind('/')
                fqname = fqname[:split_pos]
            if not fqname in pos_mark[fileorder]: continue
            for i in range(len(pos_mark[fileorder][fqname])):
                start,end = pos_mark[fileorder][fqname][i]
                s_name = fqname
                #if len(pos_mark[pos])>1:
                #    s_name += '_'+str(i)
                s_read = reads[start:end]
                if len(s_read)<filter:
                    continue
                s_qua = quality[start:end]
                if report_clip:
                    s_read_begin = reads[:start]
                    s_read_end = reads[end:]
                    s_qua_begin = quality[:start]
                    s_qua_end = quality[:end]
                leftdel = start
                rightdel = len(reads)-end
                if rightdel<0:
                    rightdel=0
                s_final = '@'+s_name+'_'+str(i)+'_'+str(fileorder)+'_'+str(leftdel)+'_'+str(rightdel)+'\n'+s_read+'\n'+'+\n'+s_qua+'\n'
                s_final = s_final.encode()
                if report_clip:
                    s_begin = '@'+s_name+'_'+str(fileorder)+'_'+str(leftdel)+'_'+str(rightdel)+'\n'+s_read_begin+'\n'+'+\n'+s_qua_begin+'\n'
                    s_end = '@'+s_name+'_'+str(fileorder)+'_'+str(leftdel)+'_'+str(rightdel)+'\n'+s_read_end+'\n'+'+\n'+s_qua_end+'\n'
                    s_begin, s_end = s_begin.encode(), s_end.encode()
                    if len(s_read_begin)>0:
                        result_begin.append(s_begin)
                    if len(s_read_end)>0:
                        result_end.append(s_end)
                result.append(s_final)
            if len(result)>2000000:
                with gzip.open(outputname+'_finalfastq.fastq.gz','a') as ff:     
                    ff.writelines(result)
                if report_clip:
                    with gzip.open(outputname+'_finalfastq_clipped_head.fastq.gz','a') as ff:
                        ff.writelines(result_begin)
                    with gzip.open(outputname+'_finalfastq_clipped_tail.fastq.gz','a') as ff:
                        ff.writelines(result_end)
                    result_begin=[]
                    result_end=[]
                result=[]
                
        fileorder+=1
        f.close()
    # print(len(result))
    if len(result)>0:
        with gzip.open(outputname+'_finalfastq.fastq.gz','a') as ff:
            ff.writelines(result)
        if report_clip:
            with gzip.open(outputname+'_finalfastq_clipped_head.fastq.gz','a') as ff:
                ff.writelines(result_begin)
            with gzip.open(outputname+'_finalfastq_clipped_tail.fastq.gz','a') as ff:
                ff.writelines(result_end)


def combine(read1,read2,strand,step):
    read_length = read2.reference_length
    tail_length = ((read_length-1)%step)+1
    if strand=='++' or strand=='--':
        refseq = read2.get_tag("XR")[-2-tail_length:-2]
        add_seq = read2.seq[-1*tail_length:]
        add_qua = read2.query_qualities[-1*tail_length:]
        read1.query_qualities.extend(add_qua)
        read1_qua = read1.query_qualities
        read1.seq = read1.seq + add_seq
        read1.query_qualities = read1_qua
    else:
        refseq = read2.get_tag("XR")[2:2+tail_length]
        add_seq = read2.seq[:tail_length]
        add_qua = read2.query_qualities[:tail_length]
        add_qua.extend(read1.query_qualities)
        read1.seq = add_seq + read1.seq
        read1.query_qualities = add_qua
        
    read1.reference_start = min(read1.reference_start,read2.reference_start)
    if read2.get_tag("NM")>0:
        new_mismatch = read1.get_tag("NM")
        for i in range(len(refseq)):
            if refseq[i]==add_seq[i]: continue
            if strand[0]=='+':
                if (refseq[i]=='C' and add_seq[i]=='T'): continue
            else:
                if (refseq[i]=='G' and add_seq[i]=='A'): continue
            new_mismatch+=1
        read1.set_tag("NM",new_mismatch)


def fragCombine(param):#,outf):
    frags = param['frags']
    order = param['order']
    args = param['args']
    step = args['step']
    filter = args['filter']
    window = args['binsize']
    order_max_gap = window//step
    #print(len(frags),order)
    strand = []
    mismatch = []
    for fr in frags:
        strand.append(fr.get_tag("ZS"))
        mismatch.append(fr.get_tag("NM"))
        
    head_frag_length_diff = 0
    if order[0]==0:
        head_frag_length_diff = abs(window - frags[0].reference_length)

    result_fragment = [frags[0]]
    result_order = [order[0]]
    result_strand = [strand[0]]
    result_extension = [0]
    for i in range(1,len(frags)):
        join_marker = False
        candidate = frags[i]
        candidate_order = order[i]
        for j, read in enumerate(result_fragment):
            if read.reference_name==candidate.reference_name and strand[i]==result_strand[j]:
                read_order = result_order[j]
                order_diff = candidate_order - read_order
                if strand[i]=='+-' or strand[i]=='-+':
                    order_diff = order_diff - result_extension[j]
                if order_diff>=order_max_gap: continue
                pos_diff = abs(read.reference_start - candidate.reference_start)
                #print(candidate_order,read_order,result_extension[j],pos_diff, order_diff)
                #print(result_extension)
                if order_diff*step==pos_diff or (j==0 and order_diff*step==pos_diff+head_frag_length_diff):
                    join_marker = True
                    combine(read, candidate,result_strand[j],step)
                    update_read(read)
                    result_extension[j] += 1
        if not join_marker:
            result_fragment.append(frags[i])
            result_order.append(order[i])
            result_strand.append(strand[i])
            result_extension.append(0)

    # Finish combining fragments. Now apply filters towards combined fragments:
    # 1. 
    valid_frags = []
    for i, frag in enumerate(result_fragment):
        #print(frag.to_string())
        length = len(frag.seq)
        if length>=filter:# and frag.get_tag("NM")<=length*0.05:
            valid_frags.append([frag,result_order[i],result_extension[i]])

    result_fragment = valid_frags
    valid_frags = []
    del_mark = [0] * len(result_fragment)
    frag_num = len(result_fragment)
    for i in range(frag_num):
        for j in range(i+1,frag_num):
            read1, order1, ext1 = result_fragment[i]
            read2, order2, ext2 = result_fragment[j]
            gap = order2 - order1
            if gap<ext1+order_max_gap:
                ext_diff = ext1-ext2
                if ext_diff>0:
                    del_mark[j]=1
                elif ext_diff<0: del_mark[i]=1
                else:
                    mis_diff = read1.get_tag("NM") - read2.get_tag("NM")
                    if mis_diff>0: del_mark[i]=1
                    else: del_mark[j]=1

    read_length = []
    for i,frag in enumerate(result_fragment):
        if del_mark[i]==0:
            valid_frags.append(result_fragment[i])
            read_length.append(result_fragment[i][0].reference_length)

    frag_rank = 0

    combined_frags = []
    for i in range(len(valid_frags)):
        frag, order, ext = valid_frags[i]
        #frag.query_name = frag.query_name + '_' + str(frag_rank) + '_' + str(order*step) + '_' + str(window+(order+ext)*step)
        name = frag.query_name
        if '_' in name[-5:]:
            name = frag.query_name.split('_')
            frag.query_name = name[0]
            frag.set_tag("MT", name[1])
        else:
            frag.set_tag("MT", '3')
        frag.set_tag("FG", str(frag_rank))
        frag.set_tag("LC", str(order*step))
        frag.set_tag("RC", str(window+(order+ext)*step))
        frag.set_tag('XR', None)
        # print(frag.to_string())
        #outf.write(frag)
        combined_frags.append(frag)
        frag_rank += 1

    return len(combined_frags), read_length, combined_frags 


def pairwise(combined_frags):
    for m1_frag in combined_frags[0]:
        for m2_frag in combined_frags[1]:
            r1 = m1_frag
            r1_mate = r1.get_tag('MT')
            r1_frag = r1.get_tag('FG')
            r1_chr = r1.reference_name
            r1_start = r1.reference_start
            r1_len = r1.reference_length
            r1_end = r1_start + r1_len
            r2 = m2_frag
            r2_mate = r2.get_tag('MT')
            r2_frag = r2.get_tag('FG')
            r2_chr = r2.reference_name
            r2_start = r2.reference_start
            r2_len = r2.reference_length
            r2_end = r2_start + r2_len
            if r1_chr==r2_chr:
                template_length = max(r1_end, r2_end) - min(r1_start, r2_start)
                if template_length<r1.template_length or r1.next_reference_start<0 or r1.next_reference_name!=r1_chr:
                    r1.template_length = template_length
                    r1.next_reference_name = r2_chr
                    r1.next_reference_start = r2_start
                    if r1_mate==1:
                        m_info=64
                    else:
                        m_info=128
                    if r2.flag&16>0:
                        r1.flag = r1.flag|32
                    r1.flag = r1.flag|1
                    r1.flag = r1.flag|m_info
                if template_length<r2.template_length or r2.next_reference_start<0 or r2.next_reference_name!=r2_chr:
                    r2.template_length = template_length
                    r2.next_reference_name = r1_chr
                    r2.next_reference_start = r1_start
                    if r2_mate==1:
                        m_info=64
                    else:
                        m_info=128
                    if r1.flag&16>0:
                        r2.flag = r2.flag|32
                    r2.flag = r2.flag|1
                    r2.flag = r2.flag|m_info
            else:
                if r1.next_reference_start<0:
                    r1.next_reference_name = r2_chr
                    r1.next_reference_start = r2_start
                    if r1_mate==1:
                        m_info=64
                    else:
                        m_info=128
                    if r2.flag&16>0:
                        r1.flag = r1.flag|32
                    r1.flag = r1.flag|1
                    r1.flag = r1.flag|m_info
                if r2.next_reference_start<0:
                    r2.next_reference_name = r1_chr
                    r2.next_reference_start = r1_start
                    if r2_mate==1:
                        m_info=64
                    else:
                        m_info=128
                    if r1.flag&16>0:
                        r2.flag = r2.flag|32
                    r2.flag = r2.flag|1
                    r2.flag = r2.flag|m_info



def unsortedCombine(unmapped_file,args):
    step = args['step']
    outputname = args['outputname']
    window = args['binsize']
    filter = args['filter']
    inputfilenumber = len(args['originalfile'])
    least_frags = (filter-window)//step

    result_fragment = []

    feed_in_parameter = {
        'args' : args,
    }

    reads_num_dist = [0]*10
    reads_len_dist = [0]*200

    with pysam.AlignmentFile(unmapped_file+'.bam','r') as f:
        wf = pysam.AlignmentFile(outputname+"_split.bam", "wb", template=f)
        frags = [[],[]]
        order = [[],[]]
        current_reads_name=''
        for line in f:
            mismatch = line.get_tag('NM')
            if mismatch>line.query_length*0.05: continue
            #print(line.query_name)
            if '&' not in line.query_name[-5:]:
                continue
            read_name, o = line.query_name.split('&')

            line.query_name = read_name
            if inputfilenumber>1:#'_' in line.query_name[-5:]:
                c = line.query_name.split('_')
                pure_name = c[0]
                _mate = int(c[-1])-1
            else:
                pure_name = line.query_name
                _mate = 0

            #if 'E00488:423:HYHFMCCXY:8:1101:14874:1872' not in read_name: continue
            #print(line.to_string())
            #if read_name!=current_reads_name:
            if pure_name!=current_reads_name:
                #print(current_reads_name)
                if current_reads_name!='':
                    combined_frags = [[],[]]
                    
                    # Combined fragments retrieving
                    for m in range(2):
                        if len(frags[m])>least_frags:
                            feed_in_parameter['frags'] = frags[m]
                            feed_in_parameter['order'] = order[m]
                            reads_num, reads_len, combined_frags[m] = fragCombine(feed_in_parameter)#fragCombine(feed_in_parameter,wf)
                            if reads_num>0:
                                reads_num_dist[reads_num] += 1
                                for l in reads_len:
                                    reads_len_dist[l] += 1
                    #Combined fragments pairing
                    if len(combined_frags[0])>0 and len(combined_frags[1])>0:
                        pairwise(combined_frags)
                    
                    #Combined fragments output
                    for m in range(2):
                        for outfrag in combined_frags[m]:
                            wf.write(outfrag)

                    frags = [[],[]]
                    order = [[],[]]
                frags[_mate].append(line)
                order[_mate].append(int(o))
                current_reads_name = pure_name # read_name
            else:
                frags[_mate].append(line)
                order[_mate].append(int(o))
    if len(frags)>least_frags:
        combined_frags = [[],[]]
        for m in range(2):
            if len(frags[m])>least_frags:
                feed_in_parameter['frags'] = frags[m]
                feed_in_parameter['order'] = order[m]
                reads_num, reads_len, combined_frags[m] = fragCombine(feed_in_parameter)
                if reads_num>0:
                    reads_num_dist[reads_num] += 1
                    for l in reads_len:
                        reads_len_dist[l] += 1
        if len(combined_frags[0])>0 and len(combined_frags[1])>0:
            pairwise(combined_frags)
        for m in range(2):
            for outfrag in combined_frags[m]:
                wf.write(outfrag)
        frags = [[],[]]
        order = [[],[]]
    
    return reads_num_dist,reads_len_dist

if __name__=='__main__':

    args={'step':5,
          'binsize':40,
          'filter':46,
          'outputname':'sample1',
          #'originalfile':['mate1.fq.gz','mate2.fq.gz'],
          #'mapfilenumber':10,
          #'finish':1,
          #'report_clip':1
    }
    #unsortedCombine('/data/yyin/LiBis_test/data/simulation/rht_1sample/LiBis_011_w40_unmap/1_1.unmapped',args)
    unsortedCombine('/scratch0/MB4CSF/MB4.R1.fq.ae.unmapped', args)
    #unsortedCombine('test_mr', args)

'''
    mr_file = ['mate1_bsmap_0.mapreduce',
               'mate1_bsmap_1.mapreduce',
               'mate1_bsmap_2.mapreduce',
               'mate1_bsmap_3.mapreduce',
               'mate1_bsmap_4.mapreduce',
               'mate1_bsmap_5.mapreduce',
               'mate1_bsmap_6.mapreduce',
               'mate1_bsmap_7.mapreduce',
               'mate1_bsmap_8.mapreduce',
               'mate1_bsmap_9.mapreduce']
    names = reads_map('mate1_bsmap.unmapped.fastq',args)
    #print(names)
    # reads_reduce(mr_file,args)
    reads_reduce(names, args)
#    step = args['step']
##    length_bin = args['binsize']
#    filter = args['filter']
#    outputname = args['outputname']
#    originalfile = args['originalfile']
'''



'''
Tested
'''
from .utils import *
import os
class TRIM():
    
    def check(self,nocheck=False):
        if not toolcheck('trim_galore -v'):
            return False,'trim_galore not found!'
        if os.path.exists('Trim'):
            if nocheck:
                print('"Trim" exist! But --nocheck enabled, so continue running')
            else:
                return False,'"Trim" file/dic exist! Please delete it.'
        else:
            os.mkdir('Trim')
        return True,''

    def setpath(self,path):
        self.path = path+'Trim'

    def run(self,filename,pair):
        if pair==1:
            p = Pshell('trim_galore --gz -o '+self.path+' '+filename)
        else:
            p = Pshell('trim_galore --gz --paired -o '+self.path+' '+filename)
        p.process()

Trim = TRIM()

if __name__=='__main__':
    tr = Trim()
    #tr.check()
    tr.setpath('./')
    tr.run('../trimtest/SRR1248444_1.1.1.1.fastq')





'''
Tested
'''
import os
from .utils import *

class BEDTOOLS:

    def check(self,nockeck=False):
        if not toolcheck('bedtools --version'):
            return False,'Bedtools not found!'
        return True,''

    def setparam(self,param):
        self.genome = param['genome']
        if 'bin' in param and param['bin']!=None:
            self.bin = param['bin']
        else:
            self.bin=1000000

    def makewindow(self):
        path=os.path.abspath(__file__)
        path = path[:path.rfind('/')+1]
        filename = path+'chromsize/'+self.genome + '.chrom.sizes'
        #print(filename)
        outputname = self.genome+'_'+str(self.bin)+'.bed'
        os.system('bedtools makewindows -g '+filename+' -w '+str(self.bin)+' > '+outputname)
        self.binfile = outputname

    def intersect(self,names):
        sample=0
        result=[]
        for name in names:
            temp = str(sample)+'.intersect'
            output = str(sample)+'.bed'
            os.system('bedtools intersect -loj -a '+self.binfile+' -b '+ name +" | awk -v OFS='\t' '{if ($7>=0 && $7<=1) print $1,$2,$3,$7}' > BED_FILE/"+temp)
            os.system('bedtools groupby -i BED_FILE/'+temp+' -g 1,2,3 -c 4 -o mean > '+'BED_FILE/'+output)
            result.append('BED_FILE/'+output)
            sample=sample+1
        return result

    '''
    Two plots from segmented genome average: TSNE/PCA and heatmap
    '''

Bedtools=BEDTOOLS()

if __name__=="__main__":
    b = Bedtools()
    param={'genome':'hg19'}
    b.setparam(param)
    print(b.check())
    b.makewindow()
    name = ['FWAC.bed']
    print(b.intersect(name))
    print(b.binfile)



import sys

try:
    from setuptools import setup
except:
    from distutils.core import setup


#!/usr/bin/env python


import sys
if sys.version_info < (3, 4):
    sys.exit('Python 3.4 or greater is required.')

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup



with open('RELEASE') as f:
    lines = f.readlines()

version = lines[0]
version = version.strip().split()[-1]

VERSION = version

LICENSE = "MIT"


setup(
    name='LiBis',
    version=VERSION,
    description=(
        'Low input Bisulfite sequencing alignment'
    ),
    long_description='',
    author='Yue Yin',
    author_email='dangertrip@tamu.edu',
    maintainer='Deqiang Sun',
    maintainer_email='dsun@tamu.edu',
    license=LICENSE,
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/Dangertrip/LiBis',
    install_requires=[
        "matplotlib",
        "numpy",
        "pandas",
        "scikit-learn",
        "scipy",
        "seaborn",
        "pysam"
    ],
    scripts=[
        "bin/LiBis",
    ],
    include_package_data=True,
)



import subprocess

def processline(line,dic,order,cov):
    if line[0]=='#': return
    t = line.strip().split()
    key=(t[0],t[1])
    if len(t)>4 and int(t[4])<cov: return
    if not key in dic:
        if order==0:
            dic[key]=[t[3]]
    else:
        dic[key].append(t[3])

def formdata(files,cov=0,bedfile=''):
    dic={}
    i=0
    for file in files:
        if bedfile=='':
            with open(file) as f:
                for line in f:
                    processline(line,dic,i,cov)
        else:
            p = subprocess.Popen('bedtools intersect -a %s -b %s' %(file,bedfile),shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
            for line in p.stdout:
                processline(line,dic,i,cov)
        i+=1
    result=[]
    for key in dic:
        if len(dic[key])==len(files):
            result.append(dic[key])
    return result

if __name__=="__main__":
    import sys
    argv = sys.argv[1:]
    final=[argv]
    final.extend(formdata(argv))
    ans=[]
    for f in final:
        s=str(f[0])
        for i in range(1,len(f)):
            s=s+','+str(f[i])
        s=s+'/n'
        ans.append(s)
    with open('data.csv','w') as f:
        f.writelines(ans)
